from wbcore.tests.conftest import *  # isort: skip # type: ignore
from django.apps import apps
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register
from wbcore.contrib.authentication.factories import (
    AuthenticatedPersonFactory,
    SuperUserFactory,
    UserFactory,
)
from wbcore.contrib.currency.factories import CurrencyFactory, CurrencyFXRatesFactory
from wbcore.contrib.directory.factories.entries import (
    CompanyFactory,
    CompanyTypeFactory,
    CustomerStatusFactory,
    EntryFactory,
    PersonFactory,
)
from wbcore.contrib.geography.factories import (
    CityFactory,
    ContinentFactory,
    CountryFactory,
    StateFactory,
)
from wbcore.contrib.geography.tests.signals import app_pre_migration
from wbcrm.factories import AccountFactory, AccountRoleFactory, AccountRoleTypeFactory
from wbfdm.factories import ExchangeFactory, InstrumentFactory, InstrumentTypeFactory, InstrumentPriceFactory
from wbportfolio.factories import (
    ClaimFactory,
    CustomerTradeFactory,
    FeesFactory,
    PortfolioFactory,
    ProductFactory,
    ProductPortfolioRoleFactory,
)

from ..factories import (
    AccountTypeRoleCommissionFactory,
    CommissionExclusionRuleFactory,
    CommissionFactory,
    CommissionRoleFactory,
    CommissionTypeFactory,
    PortfolioRoleCommissionFactory,
    RebateFactory,
)

register(AccountFactory)
register(AccountRoleFactory)
register(AccountRoleTypeFactory)

register(InstrumentFactory)
register(InstrumentTypeFactory)
register(ExchangeFactory)
register(ProductFactory)
register(ProductPortfolioRoleFactory)
register(FeesFactory)
register(PortfolioFactory)
register(InstrumentPriceFactory)
register(ClaimFactory)
register(CustomerTradeFactory)
register(CurrencyFXRatesFactory)

register(CurrencyFactory)
register(CityFactory)
register(StateFactory)
register(CountryFactory)
register(ContinentFactory)

register(CompanyFactory)
register(EntryFactory)
register(PersonFactory)
register(CustomerStatusFactory)
register(CompanyTypeFactory)

register(AuthenticatedPersonFactory, "authenticated_person")
register(UserFactory)
register(SuperUserFactory, "superuser")

register(RebateFactory)
register(CommissionTypeFactory)
register(CommissionFactory)
register(CommissionExclusionRuleFactory)
register(AccountTypeRoleCommissionFactory, "account_role_type_commission")
register(CommissionRoleFactory)
register(PortfolioRoleCommissionFactory, "portfolio_role_commission")
from .signals import *

pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("wbportfolio"))
from .signals import *  # noqa: F401



from datetime import date
from io import BytesIO

import pandas as pd
from celery import shared_task
from wbcore.contrib.authentication.models import User
from wbcore.contrib.directory.models import Entry
from wbcrm.models.accounts import Account
from wbportfolio.models.products import Product

from wbcommission.models import Commission, CommissionType, Rebate

from .utils import create_report_and_send


@shared_task(queue="commission")
def create_audit_report_and_send_as_task(user_id: int, recipient_id: int, start_date: date, end_date: date):
    user = User.objects.get(id=user_id)
    recipient = Entry.objects.get(id=recipient_id)
    create_report_and_send(user, recipient, start_date, end_date, create_report)


def create_report(user, customer, start_date, end_date):
    rebates = Rebate.objects.filter(recipient=customer, date__gte=start_date, date__lte=end_date).filter_for_user(user)
    df = pd.DataFrame(
        rebates.values(
            "date",
            "account",
            "product",
            "recipient",
            "commission_type",
            "commission",
            "value",
            "audit_log",
        )
    )
    buffer = BytesIO()
    if df.empty:
        raise ValueError("There is not rebate for this customer and given time period")
    df = pd.concat([df[df.columns.difference(["audit_log"])], pd.json_normalize(df["audit_log"])], axis=1).sort_values(
        by="date"
    )
    # we deserialize ids into human readable name
    df.commission_type = df.commission_type.map(dict(CommissionType.objects.values_list("id", "name")))
    df["product"] = df["product"].map(dict(Product.objects.values_list("id", "computed_str")))
    df.commission = df.commission.map(
        dict(map(lambda x: (x, str(Commission.objects.get(id=x))), df.commission.unique()))
    )
    df.recipient = df.recipient.map(dict(Entry.objects.values_list("id", "computed_str")))
    df.account = df.account.map(dict(Account.objects.values_list("id", "computed_str")))
    df.to_csv(buffer, index=False, mode="wb", encoding="UTF-8")
    return buffer, "audit_report_{}_{}_{}.csv".format(customer.computed_str, start_date, end_date), "application/csv"



from datetime import date, timedelta
from decimal import Decimal

import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from faker import Faker
from pandas.tseries.offsets import BDay
from psycopg.types.range import DateRange, NumericRange
from wbportfolio.models import Claim, Trade
from wbportfolio.models.roles import PortfolioRole

from wbcommission.models import Commission, CommissionRule, CommissionType

fake = Faker()


@pytest.mark.django_db
class TestCommissionModel:
    def test_init(self, commission):
        assert commission.id is not None

    def test_init_account_role_type_commission(self, account_role_type_commission):
        assert account_role_type_commission.id is not None

    def test_init_portfolio_role_commission(self, portfolio_role_commission):
        assert portfolio_role_commission.id is not None

    @pytest.mark.parametrize(
        "validity_date, is_superuser",
        [
            (fake.date_object(), True),
            (fake.date_object(), False),
        ],
    )
    def test_filter_for_user(
        self,
        user,
        commission_factory,
        commission_role_factory,
        account_factory,
        account_role_factory,
        validity_date,
        is_superuser,
    ):
        # Create a public account without any account role. The linked commission rule cannot be seen by the user
        public_account_without_role = account_factory.create(is_public=True)
        public_account_without_role_commission = commission_factory.create(  # noqa
            is_hidden=fake.pybool(), account=public_account_without_role
        )  # noqa

        recipient_commission = commission_factory.create(is_hidden=fake.pybool(), crm_recipient=user.profile.entry_ptr)

        # Create a public account with a account role for the user, the underlying commision line can be seen by the user
        public_account_with_role = account_factory.create(is_public=True)
        account_role_factory.create(account=public_account_with_role, entry=user.profile.entry_ptr)
        public_account_with_role_commission = commission_factory.create(
            is_hidden=False, account=public_account_with_role
        )  # noqa
        # any public sub accounts can be seen by the user, as any commission lines attached to these sub account but as it is hidden, it will not show
        sub_public_account_with_parent_role = account_factory.create(parent=public_account_with_role)
        sub_public_account_with_parent_role_commission = commission_factory.create(  # noqa
            is_hidden=True, account=sub_public_account_with_parent_role
        )
        # Private account without role, so cannot be seen by the user, as any commission line
        private_account_without_role = account_factory.create(is_public=False)
        private_account_without_role_commission = commission_factory.create(  # noqa
            is_hidden=False, account=private_account_without_role
        )

        # Private account with a role, so the underlying commission line can be seen by the user
        private_account_with_role = account_factory.create(is_public=False)
        account_role_factory.create(account=private_account_with_role, entry=user.profile.entry_ptr)
        unhidden_commission_with_role_on_private_account = commission_factory.create(
            is_hidden=False, account=private_account_with_role
        )  # noqa

        # Private account with a role, but expired, so the commission line cannot be seen by the user
        private_account_with_unvalid_role = account_factory.create(  # noqa
            is_public=False,
        )
        account_role_factory.create(
            account=private_account_with_unvalid_role,
            entry=user.profile.entry_ptr,
            visibility_daterange=DateRange(date.min, fake.past_date()),  # type: ignore
        )
        private_account_with_unvalid_role_commission = commission_factory.create(  # noqa
            is_hidden=False, account=private_account_with_unvalid_role
        )

        # Valid role an private account but the commission line is hidden so the user cannot see this commission line
        private_account_with_valid_role_hidden_commission = commission_factory.create(  # noqa
            is_hidden=True, account=private_account_with_role
        )  # noqa Valid account role but commission line hidden, so won't show up

        # Commission line on account without any role, but the user as direct commission role so they can see it
        commission_with_direct_role = commission_role_factory.create(person=user.profile).commission
        if is_superuser:
            user.user_permissions.add(
                Permission.objects.get(content_type__app_label="wbcommission", codename="administrate_commission")
            )
        user = get_user_model().objects.get(id=user.id)
        if is_superuser:
            assert set(Commission.objects.filter_for_user(user, validity_date)) == set(Commission.objects.all())
        else:
            assert set(Commission.objects.filter_for_user(user)) == {
                public_account_with_role_commission,
                recipient_commission,
                unhidden_commission_with_role_on_private_account,
                commission_with_direct_role,
            }

    @pytest.mark.parametrize("val_date", [fake.date_object()])
    def test_get_recipients(self, commission, product, val_date):
        assert set(commission.get_recipients(commission.account, product, val_date)) == {
            (commission.crm_recipient, Decimal(1.0))
        }

    @pytest.mark.parametrize("val_date", [fake.date_object()])
    def test_get_recipients_portfolio_role(
        self,
        portfolio_role_commission_factory,
        product,
        val_date,
        product_portfolio_role_factory,
    ):
        # Create valid portfolio roles
        valid_role1 = product_portfolio_role_factory.create(instrument=product, start=val_date)
        valid_role2 = product_portfolio_role_factory.create(
            instrument=None, start=val_date, role_type=valid_role1.role_type
        )
        valid_role3 = product_portfolio_role_factory.create(
            instrument=product, start=val_date - timedelta(days=1), end=val_date, role_type=valid_role1.role_type
        )
        # create invalid portfolio roles
        unvalid_role1 = product_portfolio_role_factory.create(  # noqa
            instrument=product, end=val_date + timedelta(days=1)
        )  # noqa
        unvalid_role2 = product_portfolio_role_factory.create(  # noqa
            instrument=product,
            start=val_date,
            role_type=PortfolioRole.RoleType.ANALYST
            if valid_role1.role_type == PortfolioRole.RoleType.PORTFOLIO_MANAGER
            else PortfolioRole.RoleType.PORTFOLIO_MANAGER,
        )  # noqa

        # create commission of type portfolio role whose type correspond to valid_role1's type
        commission = portfolio_role_commission_factory.create(portfolio_role_recipient=valid_role1.role_type)
        total_weighting = valid_role1.weighting + valid_role2.weighting + valid_role3.weighting
        # We expect all recipients who have a portfolio role of type role1 to get a share of that commission line
        res = {k: v for k, v in commission.get_recipients(commission.account, product, val_date)}
        assert res[valid_role1.person.entry_ptr] == pytest.approx(
            Decimal(valid_role1.weighting / total_weighting), rel=Decimal(1e-6)
        )
        assert res[valid_role2.person.entry_ptr] == pytest.approx(
            Decimal(valid_role2.weighting / total_weighting), rel=Decimal(1e-6)
        )
        assert res[valid_role3.person.entry_ptr] == pytest.approx(
            Decimal(valid_role3.weighting / total_weighting), rel=Decimal(1e-6)
        )

    @pytest.mark.parametrize("val_date", [fake.date_object()])
    def test_get_recipients_account_role(
        self, account_type_role_commission_factory, account_factory, account_role_factory, product, val_date
    ):
        parent_account = account_factory.create()
        valid_parent_role1 = account_role_factory.create(
            account=parent_account, weighting=0.5
        )  # parent account role but not direct account role, thus invalid
        unvalid_role1 = account_role_factory.create(  # noqa
            account=parent_account,
            visibility_daterange=DateRange(date.min, val_date),  # type: ignore
        )  # parent account role but not direct account role, thus invalid
        child_account = account_factory.create(parent=parent_account)

        # create valid account role for that child account
        valid_role1 = account_role_factory.create(
            account=child_account, role_type=valid_parent_role1.role_type, weighting=0.4
        )
        valid_role2 = account_role_factory.create(
            account=child_account, role_type=valid_parent_role1.role_type, weighting=0.1
        )
        account_role_factory.create(account=child_account)  # noqa Other role type, therefore unvalid

        # Create commission of type account type role for that accout role type of unvalid_role1
        account_role_type_commission = account_type_role_commission_factory.create(
            account=parent_account, account_role_type_recipient=valid_parent_role1.role_type
        )
        res = {k: v for k, v in account_role_type_commission.get_recipients(child_account, product, val_date)}
        # we expect every profile who have an account role fo type unvalid_role1.role_type to gain from this commission line
        assert res[valid_role1.entry] == pytest.approx(Decimal(0.4), rel=Decimal(1e-6))
        assert res[valid_role2.entry] == pytest.approx(Decimal(0.1), rel=Decimal(1e-6))
        assert res[valid_parent_role1.entry] == pytest.approx(Decimal(0.5), rel=Decimal(1e-6))

        assert (
            next(account_role_type_commission.get_recipients(parent_account, product, val_date))[0]
            == valid_parent_role1.entry
        )
        assert next(account_role_type_commission.get_recipients(parent_account, product, val_date))[
            1
        ] == pytest.approx(Decimal(0.5 / 0.5), rel=Decimal(1e-6))

    @pytest.mark.parametrize("validity_date, min_aum", [(fake.date_object(), fake.pydecimal(min_value=0))])
    def test_is_valid(self, commission, product, validity_date, min_aum):
        # basic test for is_valid. We expect the "validated_percent" and "validated_net_commission" to be set properly upon validation
        rule = commission.rules.first()
        rule.percent = Decimal(0.2)
        rule.save()
        rule.refresh_from_db()
        assert commission.is_valid(validity_date, product, min_aum)
        assert commission.validated_percent == rule.percent
        assert commission.validated_net_commission == commission.net_commission

    @pytest.mark.parametrize("validity_date, min_aum", [(fake.date_object(), fake.pydecimal(min_value=0))])
    def test_is_valid_with_exclusion_rule(
        self, commission, commission_exclusion_rule_factory, product, validity_date, min_aum, account_role_type
    ):
        # test overriding of commission rule by exclusion rule on a specific product
        rule = commission.rules.first()
        rule.percent = Decimal(0.2)
        rule.save()
        rule.refresh_from_db()
        exclusion_rule = commission_exclusion_rule_factory.create(
            product=product,
            commission_type=commission.commission_type,
        )
        exclusion_rule.refresh_from_db()
        assert commission.is_valid(validity_date, product, min_aum)
        assert commission.validated_percent == exclusion_rule.overriding_percent
        assert commission.validated_net_commission == exclusion_rule.get_net_or_gross(commission.net_commission)

        # we assign a explicit account type  for the commission
        commission.exclusion_rule_account_role_type = account_role_type
        commission.save()
        exclusion_rule_for_specific_account_type = commission_exclusion_rule_factory.create(
            product=product, commission_type=commission.commission_type, account_role_type=account_role_type
        )
        exclusion_rule_for_specific_account_type.refresh_from_db()
        assert commission.is_valid(validity_date, product, min_aum)
        assert commission.validated_percent == exclusion_rule_for_specific_account_type.overriding_percent
        assert commission.validated_net_commission == exclusion_rule_for_specific_account_type.get_net_or_gross(
            commission.net_commission
        )

    @pytest.mark.parametrize(
        "validity_date, min_aum", [(fake.date_object(), fake.pydecimal(min_value=10, max_value=100))]
    )
    def test_is_invalid_aum(self, commission, product, validity_date, min_aum):
        rule = commission.rules.first()
        rule.assets_under_management_range = NumericRange(min_aum, None)  # type: ignore
        rule.save()
        assert not commission.is_valid(validity_date, product, min_aum - Decimal(1))

    @pytest.mark.parametrize("validity_date, min_aum", [(fake.date_object(), fake.pydecimal(min_value=1000000))])
    def test_is_invalid_date(self, commission, product, validity_date, min_aum):
        rule = commission.rules.first()
        rule.timespan = DateRange(validity_date + timedelta(days=1), date.max)  # type: ignore
        rule.save()
        assert not commission.is_valid(validity_date, product, min_aum)


@pytest.mark.django_db
class TestCommissionType:
    @pytest.mark.parametrize(
        "commission_type__name,net_commission,compute_date,percent1,percent2",
        [
            (
                "MANAGEMENT",
                True,
                fake.date_object(),
                fake.pydecimal(min_value=0, max_value=1, right_digits=2) / Decimal(2),
                fake.pydecimal(min_value=0, max_value=1, right_digits=2) / Decimal(2),
            ),
            (
                "MANAGEMENT",
                False,
                fake.date_object(),
                fake.pydecimal(min_value=0, max_value=1, right_digits=2) / Decimal(2),
                fake.pydecimal(min_value=0, max_value=1, right_digits=2) / Decimal(2),
            ),
            (
                "PERFORMANCE",
                True,
                fake.date_object(),
                fake.pydecimal(min_value=0, max_value=1, right_digits=2) / Decimal(2),
                fake.pydecimal(min_value=0, max_value=1, right_digits=2) / Decimal(2),
            ),
            (
                "PERFORMANCE",
                False,
                fake.date_object(),
                fake.pydecimal(min_value=0, max_value=1, right_digits=2) / Decimal(2),
                fake.pydecimal(min_value=0, max_value=1, right_digits=2) / Decimal(2),
            ),
        ],
    )
    def test_get_valid_commissions(
        self, account, product, commission_factory, commission_type, net_commission, compute_date, percent1, percent2
    ):
        # Here we test the following:
        # - computation of net/gross fees
        # - invalid commission rules are not considered
        invalid_aum_commission = commission_factory.create(  # noqa
            account=account,
            commission_type=commission_type,
            net_commission=net_commission,
            order=0,
            rule_aum=NumericRange(1e6 + 1, None),  # type: ignore
        )  # invalid commission rule because the aum range is below the total aum
        invalid_date_commission = commission_factory.create(  # noqa
            account=account,
            commission_type=commission_type,
            net_commission=net_commission,
            order=1,
            rule_timespan=DateRange(date.min, compute_date),  # type: ignore
        )  # invalid commission rule because the rule is not valid for the given date

        commission1 = commission_factory.create(
            account=account,
            commission_type=commission_type,
            net_commission=net_commission,
            order=2,
            rule_percent=percent1,
        )
        commission2 = commission_factory.create(
            account=account,
            commission_type=commission_type,
            net_commission=net_commission,
            order=3,
            rule_percent=percent2,
        )
        commission3 = commission_factory.create(
            account=account,
            commission_type=commission_type,
            net_commission=False,
            order=4,
            rule_percent=Decimal(1) - percent1 - percent2 + Decimal(0.5),
        )  # We add to much percent so that the sum of gross percent is greater than 1
        commission4 = commission_factory.create(  # noqa
            account=account, commission_type=commission_type, net_commission=False, order=5, rule_percent=percent1
        )  # there isn't any gross percent left, so the result is always zero

        res = list(commission_type.get_valid_commissions(account, compute_date, product, Decimal(1e6)))
        if net_commission:
            assert res == [
                (commission1, percent1),
                (commission2, (Decimal(1.0) - percent1) * percent2),
                (commission3, Decimal(1) - percent1 - (Decimal(1) - percent1) * percent2),
            ]
        else:
            if (
                Decimal(1) - percent1 - percent2 == 0
            ):  # in that case we don't exepct commission 3 to show up as a valid commission as the resulting percent will be zero (i.e. no percent left to assign)
                assert res == [(commission1, percent1), (commission2, percent2)]
            else:
                assert res == [
                    (commission1, percent1),
                    (commission2, percent2),
                    (commission3, Decimal(1) - percent1 - percent2),
                ]

    def test_get_valid_commissions_with_inheritance(
        self, account_factory, commission_factory, commission_type, product
    ):
        compute_date = fake.date_object()
        parent_account = account_factory.create()
        account = account_factory.create(parent=parent_account)

        parent_commission_0 = commission_factory.create(
            account=parent_account,
            commission_type=commission_type,
            net_commission=False,
            order=0,
            rule_percent=Decimal(0.1),
        )
        parent_commission_1 = commission_factory.create(  # noqa
            account=parent_account,
            commission_type=commission_type,
            net_commission=False,
            order=1,
            rule_percent=Decimal(0.2),
        )
        child_account_1 = commission_factory.create(
            account=account, commission_type=commission_type, net_commission=False, order=1, rule_percent=Decimal(0.3)
        )
        child_account_2 = commission_factory.create(
            account=account, commission_type=commission_type, net_commission=False, order=2, rule_percent=Decimal(0.4)
        )
        res = list(commission_type.get_valid_commissions(account, compute_date, product, Decimal(1e6)))
        assert res == [
            (parent_commission_0, parent_commission_0.rules.first().percent),
            (child_account_1, child_account_1.rules.first().percent),
            (child_account_2, child_account_2.rules.first().percent),
        ]

    @pytest.mark.parametrize(
        "val_date, percent1,percent2",
        [
            (
                fake.date_object(),
                fake.pydecimal(positive=True, max_value=1, right_digits=2) / Decimal(2),
                fake.pydecimal(positive=True, max_value=1, right_digits=2) / Decimal(2),
            )
        ],
    )
    def test_compute_rebates(
        self,
        val_date,
        product,
        account,
        fees_factory,
        customer_trade_factory,
        commission_factory,
        account_type_role_commission_factory,
        account_role_factory,
        instrument_price_factory,
        claim_factory,
        percent1,
        percent2,
    ):
        val_date = (val_date + BDay(0)).date()
        val_date_1 = (val_date - BDay(1)).date()
        fees_factory.create(linked_product=product, transaction_date=val_date_1, transaction_subtype="PERFORMANCE")
        fees_factory.create(linked_product=product, transaction_date=val_date_1, transaction_subtype="MANAGEMENT")
        perf_fees = fees_factory.create(
            linked_product=product, transaction_date=val_date, transaction_subtype="PERFORMANCE"
        )
        mngt_fees = fees_factory.create(
            linked_product=product, transaction_date=val_date, transaction_subtype="MANAGEMENT"
        )
        sub2 = customer_trade_factory.create(
            underlying_instrument=product,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            value_date=val_date,
            transaction_date=val_date_1,  # we consider only trade in t-1
        )
        sub1 = customer_trade_factory.create(
            underlying_instrument=product,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            value_date=val_date,
            transaction_date=val_date_1,
        )

        claim = claim_factory(
            trade__value_date=val_date,
            trade__transaction_date=val_date_1,
            account=account,
            trade=sub1,
            product=sub1.underlying_instrument,
            status=Claim.Status.APPROVED,
        )

        product_shares = sub1.shares + sub2.shares
        instrument_price_factory.create(
            instrument=product, outstanding_shares=product_shares, date=val_date, calculated=True
        )
        instrument_price_factory.create(
            instrument=product, outstanding_shares=product_shares, date=val_date, calculated=False
        )
        instrument_price_factory.create(
            instrument=product, outstanding_shares=product_shares, date=val_date_1, calculated=True
        )
        instrument_price_factory.create(
            instrument=product, outstanding_shares=product_shares, date=val_date_1, calculated=False
        )
        account_role1 = account_role_factory.create(account=account)
        account_role2 = account_role_factory.create(account=account, role_type=account_role1.role_type)

        commission_perf = commission_factory.create(
            account=account, commission_type__name="PERFORMANCE", order=0, rule_percent=percent1
        )
        commission_perf_account_role = account_type_role_commission_factory.create(  # noqa
            account=account,
            commission_type__name="PERFORMANCE",
            order=1,
            rule_percent=percent2,
            account_role_type_recipient=account_role1.role_type,
        )
        commission_mngt = commission_factory.create(
            account=account, commission_type__name="MANAGEMENT", order=0, rule_percent=percent1
        )
        res = dict()
        for commission_type in CommissionType.objects.all():
            for (
                terminal_account,
                compute_date,
                commission,
                product,
                recipient,
                recipient_fees,
                _,
            ) in commission_type.compute_rebates(account):
                res[recipient] = recipient_fees
        assert res[commission_perf.crm_recipient] == pytest.approx(
            perf_fees.total_value * (claim.shares / product_shares) * percent1, rel=Decimal(1e-4)
        )
        assert res[account_role1.entry] == pytest.approx(
            perf_fees.total_value
            * (claim.shares / product_shares)
            * (Decimal(1.0) - percent1)
            * percent2
            / Decimal(2),
            rel=Decimal(1e-4),
        )
        assert res[account_role2.entry] == pytest.approx(
            perf_fees.total_value
            * (claim.shares / product_shares)
            * (Decimal(1.0) - percent1)
            * percent2
            / Decimal(2),
            rel=Decimal(1e-4),
        )
        assert res[commission_mngt.crm_recipient] == pytest.approx(
            mngt_fees.total_value * (claim.shares / product_shares) * percent1, rel=Decimal(1e-4)
        )

    def test_account_merging(self, account_factory, commission_factory, commission_role_factory, rebate_factory):
        # TODO implemetns for commission
        pivot_date = date(2023, 1, 1)
        # for each type a
        base_account = account_factory.create()
        merged_account = account_factory.create()

        base_mngt_commission = commission_factory.create(
            account=base_account, rule_percent=0.2, rule_timespan=DateRange(date.min, pivot_date)
        )
        base_rule = base_mngt_commission.rules.first()
        base_mngt_commission_role = commission_role_factory.create(commission=base_mngt_commission)
        base_mngt_rebate = rebate_factory.create(commission=base_mngt_commission, account=base_account)

        merged_mngt_commission = commission_factory.create(
            crm_recipient=base_mngt_commission.crm_recipient,
            account=merged_account,
            rule_percent=0.5,
            rule_timespan=DateRange(pivot_date + timedelta(days=1), date.max),
        )
        CommissionRule.objects.create(  # we create a rule that overlaps the base rule but to be sure it does not change the base commission rule
            commission=merged_mngt_commission,
            timespan=DateRange(date.min, pivot_date),  # type: ignore
            assets_under_management_range=NumericRange(0, 1000000000000),  # type: ignore
        )
        merged_mngt_commission_role = commission_role_factory.create(commission=merged_mngt_commission)
        merged_mngt_rebate = rebate_factory.create(commission=merged_mngt_commission, account=merged_account)
        merged_mngt_commission_rule = merged_mngt_commission.rules.first()

        merged_perf_commission = commission_factory.create(
            crm_recipient=base_mngt_commission.crm_recipient,
            account=merged_account,
            commission_type__name="Performance",
        )
        merged_perf_commission_rule = merged_perf_commission.rules.first()
        merged_perf_commission_role = commission_role_factory.create(commission=merged_perf_commission)
        merged_perf_rebate = rebate_factory.create(commission=merged_perf_commission, account=merged_account)

        base_account.merge(merged_account)

        base_mngt_commission.refresh_from_db()

        # test that the non overlapping rule are forwarded to the existing commission
        assert set(base_mngt_commission.rules.all()) == {base_rule, merged_mngt_commission_rule}
        assert set(base_mngt_commission.roles.all()) == {base_mngt_commission_role, merged_mngt_commission_role}

        # test that the base rebate didn't change
        base_mngt_rebate.refresh_from_db()
        assert base_mngt_rebate.commission == base_mngt_commission
        assert base_mngt_rebate.account == base_account

        # test that the overlapping commission are simply deleted
        with pytest.raises(Commission.DoesNotExist):
            merged_mngt_commission.refresh_from_db()

        # checked that the rebate from the merged commission is reassigned to the base commission and account
        merged_mngt_rebate.refresh_from_db()
        assert merged_mngt_rebate.commission == base_mngt_commission
        assert merged_mngt_rebate.account == base_account

        base_perf_commission = Commission.objects.get(
            commission_type__key="performance",
            account=base_account,
            crm_recipient=merged_perf_commission.crm_recipient,
        )
        assert set(base_perf_commission.rules.all()) == {merged_perf_commission_rule}
        assert set(base_perf_commission.roles.all()) == {merged_perf_commission_role}
        with pytest.raises(Commission.DoesNotExist):
            merged_perf_commission.refresh_from_db()
        merged_perf_rebate.refresh_from_db()
        assert merged_perf_rebate.commission == base_perf_commission
        assert merged_perf_rebate.account == base_account



import pytest
from faker import Faker
from pandas.tseries.offsets import BDay
from rest_framework.reverse import reverse
from rest_framework.test import APIClient, APIRequestFactory
from wbfdm.factories import InstrumentPriceFactory
from wbportfolio.factories import FeesFactory, ProductFactory
from wbportfolio.models import Product

from wbcommission.analytics.marginality import MarginalityCalculator
from wbcommission.factories import CommissionTypeFactory, RebateFactory
from wbcommission.viewsets.rebate import RebateProductMarginalityViewSet

fake = Faker()


def _create_fixture(product, val_date, net_value=100, outstanding_shares=100):
    management = CommissionTypeFactory.create(key="management")
    perforance = CommissionTypeFactory.create(key="performance")
    i1 = InstrumentPriceFactory.create(
        instrument=product, net_value=net_value, outstanding_shares=outstanding_shares, calculated=False, date=val_date
    ).net_value  # create a price of AUM 100*100
    management_fees_1 = FeesFactory.create(
        linked_product=product, transaction_date=val_date, transaction_subtype="MANAGEMENT", calculated=False
    ).total_value
    performance_fees_1 = FeesFactory.create(
        linked_product=product, transaction_date=val_date, transaction_subtype="PERFORMANCE", calculated=False
    ).total_value
    performance_crys_fees_1 = FeesFactory.create(
        linked_product=product,
        transaction_date=val_date,
        transaction_subtype="PERFORMANCE_CRYSTALIZED",
        calculated=False,
    ).total_value
    management_rebate_1 = RebateFactory.create(product=product, date=val_date, commission_type=management).value
    performance_rebate_1 = RebateFactory.create(product=product, date=val_date, commission_type=perforance).value
    if val_date.weekday() == 0:
        return (
            i1,
            management_fees_1 / 3.0,
            performance_fees_1 / 3.0,
            performance_crys_fees_1 / 3.0,
            float(management_rebate_1) / 3.0,
            float(performance_rebate_1) / 3.0,
        )
    else:
        return (
            i1,
            management_fees_1,
            performance_fees_1,
            performance_crys_fees_1,
            float(management_rebate_1),
            float(performance_rebate_1),
        )


@pytest.mark.django_db
class TestMarginalityCalculator:
    @pytest.fixture
    def marginality_calculator(self):
        product = ProductFactory.create()
        start = (fake.date_object() + BDay(0)).date()
        end = (start + BDay(1)).date()

        (
            self.i1,
            self.management_fees_1,
            self.performance_fees_1,
            self.performance_crys_fees_1,
            self.management_rebate_1,
            self.performance_rebate_1,
        ) = _create_fixture(product, start)
        self.product_id = product.id
        self.start = start
        self.end = end
        return MarginalityCalculator(Product.objects.filter(id=product.id), start, end)

    def test_fees(self, marginality_calculator):
        assert marginality_calculator.management_fees.loc[self.product_id] == pytest.approx(
            self.management_fees_1, rel=1e-4
        )
        assert marginality_calculator.performance_fees.loc[self.product_id] == pytest.approx(
            self.performance_fees_1 + self.performance_crys_fees_1, rel=1e-4
        )
        assert marginality_calculator.total_fees.loc[self.product_id] == pytest.approx(
            self.performance_fees_1 + self.performance_crys_fees_1 + self.management_fees_1, rel=1e-4
        )

    def test_rebates(self, marginality_calculator):
        assert marginality_calculator.management_rebates.loc[self.product_id] == pytest.approx(
            self.management_rebate_1, rel=1e-4
        )
        assert marginality_calculator.performance_rebates.loc[self.product_id] == pytest.approx(
            self.performance_rebate_1, rel=1e-4
        )
        assert marginality_calculator.total_rebates.loc[self.product_id] == pytest.approx(
            self.performance_rebate_1 + self.management_rebate_1, rel=1e-4
        )

    def test_fees_usd(self, marginality_calculator):
        assert marginality_calculator.management_fees_usd.loc[self.product_id] == pytest.approx(
            self.management_fees_1, rel=1e-4
        )
        assert marginality_calculator.performance_fees_usd.loc[self.product_id] == pytest.approx(
            self.performance_fees_1 + self.performance_crys_fees_1, rel=1e-4
        )
        assert marginality_calculator.total_fees_usd.loc[self.product_id] == pytest.approx(
            self.performance_fees_1 + self.performance_crys_fees_1 + self.management_fees_1, rel=1e-4
        )

    def test_rebates_usd(self, marginality_calculator):
        assert marginality_calculator.management_rebates_usd.loc[self.product_id] == pytest.approx(
            self.management_rebate_1, rel=1e-4
        )
        assert marginality_calculator.performance_rebates_usd.loc[self.product_id] == pytest.approx(
            self.performance_rebate_1, rel=1e-4
        )
        assert marginality_calculator.total_rebates_usd.loc[self.product_id] == pytest.approx(
            self.performance_rebate_1 + self.management_rebate_1, rel=1e-4
        )

    def test_marginality(self, marginality_calculator):
        assert marginality_calculator.management_marginality.loc[self.product_id] == pytest.approx(
            self.management_fees_1 - self.management_rebate_1, rel=1e-4
        )
        assert marginality_calculator.performance_marginality.loc[self.product_id] == pytest.approx(
            self.performance_fees_1 + self.performance_crys_fees_1 - self.performance_rebate_1, rel=1e-4
        )
        assert marginality_calculator.total_marginality.loc[self.product_id] == pytest.approx(
            (self.performance_fees_1 + self.performance_crys_fees_1 + self.management_fees_1)
            - (self.performance_rebate_1 + self.management_rebate_1),
            rel=1e-4,
        )

    def test_marginality_usd(self, marginality_calculator):
        assert marginality_calculator.management_marginality_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.management_marginality.loc[self.product_id], rel=1e-4
        )
        assert marginality_calculator.performance_marginality_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.performance_marginality.loc[self.product_id], rel=1e-4
        )
        assert marginality_calculator.total_marginality_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.total_marginality.loc[self.product_id], rel=1e-4
        )

    def test_marginality_percent(self, marginality_calculator):
        assert marginality_calculator.management_marginality_percent.loc[self.product_id] == pytest.approx(
            (self.management_fees_1 - float(self.management_rebate_1)) / self.management_fees_1, rel=1e-4
        )
        assert marginality_calculator.performance_marginality_percent.loc[self.product_id] == pytest.approx(
            (self.performance_fees_1 + self.performance_crys_fees_1 - float(self.performance_rebate_1))
            / (self.performance_fees_1 + self.performance_crys_fees_1),
            rel=1e-4,
        )
        assert marginality_calculator.total_marginality_percent.loc[self.product_id] == pytest.approx(
            (
                (self.performance_fees_1 + self.performance_crys_fees_1 + self.management_fees_1)
                - (float(self.performance_rebate_1) + float(self.management_rebate_1))
            )
            / (self.performance_fees_1 + self.performance_crys_fees_1 + self.management_fees_1),
            rel=1e-4,
        )

    def test_marginality_percent_usd(self, marginality_calculator):
        assert marginality_calculator.management_marginality_percent_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.management_marginality_percent.loc[self.product_id], rel=1e-4
        )
        assert marginality_calculator.performance_marginality_percent_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.performance_marginality_percent.loc[self.product_id], rel=1e-4
        )
        assert marginality_calculator.total_marginality_percent_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.total_marginality_percent.loc[self.product_id], rel=1e-4
        )

    def test_get_net_marginality(self, marginality_calculator):
        assert marginality_calculator.get_net_marginality("management").loc[self.product_id] == pytest.approx(
            (self.management_fees_1 - float(self.management_rebate_1)) / (100 * 100) * 360, rel=1e-4
        )

    def test_get_aggregated_net_marginality(self, marginality_calculator):
        product_1 = ProductFactory.create()
        product_2 = ProductFactory.create()
        start = (fake.date_object() + BDay(0)).date()
        end = (start + BDay(1)).date()

        (
            i1,
            management_fees_1,
            performance_fees_1,
            performance_crys_fees_1,
            management_rebate_1,
            performance_rebate_1,
        ) = _create_fixture(product_1, start)
        (
            i2,
            management_fees_2,
            performance_fees_2,
            performance_crys_fees_2,
            management_rebate_2,
            performance_rebate_2,
        ) = _create_fixture(product_2, start, outstanding_shares=1000)
        total_aum = 100 * 100 + 100 * 1000
        calculator = MarginalityCalculator(Product.objects.filter(id__in=[product_1.id, product_2.id]), start, end)
        assert calculator.get_aggregated_net_marginality("management") == pytest.approx(
            ((management_fees_1 + management_fees_2) - (management_rebate_1 + management_rebate_2)) / total_aum * 360,
            rel=1e-4,
        )

    def test_rebate_marginality_view(self, marginality_calculator, super_user):
        request = APIRequestFactory().get("")
        request.query_params = {}
        request.GET = {"date_gte": self.start.strftime("%Y-%m-%d"), "date_lte": self.end.strftime("%Y-%m-%d")}
        request.user = super_user
        viewset = RebateProductMarginalityViewSet(request=request)
        assert set(viewset._get_dataframe().columns) == {
            "id",
            "title",
            "currency_symbol",
            "base_management_fees_percent",
            "management_fees",
            "management_rebates",
            "management_marginality",
            "management_marginality_percent",
            "base_performance_fees_percent",
            "performance_fees",
            "performance_rebates",
            "performance_marginality",
            "total_fees",
            "total_rebates",
            "total_marginality_percent",
            "total_fees_usd",
            "total_rebates_usd",
            "total_marginality_usd",
            "net_management_marginality",
            "net_performance_marginality",
        }

        # regression tests to check that without date range the view doesn't break
        url = reverse("wbcommission:rebatemarginalitytable-list", args=[])
        api_client = APIClient()
        api_client.force_authenticate(super_user)

        response = api_client.options(url)
        assert response.status_code == 200

        response = api_client.get(url)
        assert response.status_code == 200

        # check that with proper date range the viewset return some content
        response = api_client.options(url, data=request.GET)
        assert response.status_code == 200

        response = api_client.get(url, data=request.GET)
        assert response.status_code == 200
        assert response.json()["results"]



# @admin.register(Rebate)
# class RebateModelAdmin(admin.ModelAdmin):
#     def has_add_permission(self, request, obj=None):
#         return False

#     def has_delete_permission(self, request, obj=None):
#         return False



from contextlib import suppress
from datetime import date

from django.core.mail import EmailMultiAlternatives
from django.template.loader import get_template
from wbcore.contrib.authentication.models import User
from wbcore.contrib.directory.models import Entry
from wbcore.utils.html import convert_html2text

AccountCustomer = None


def create_report_and_send(user: User, customer: Entry, start_date: date, end_date: date, report_callback):
    with suppress(ValueError):
        report_stream, file_name, file_extension = report_callback(user, customer, start_date, end_date)

        html = get_template("portfolio/email/customer_report.html")

        context = {"profile": user, "customer": customer}
        html_content = html.render(context)

        msg = EmailMultiAlternatives("Report", body=convert_html2text(html_content), to=[user.email])
        msg.attach_alternative(html_content, "text/html")
        report_stream.seek(0)
        msg.attach(
            file_name,
            report_stream.read(),
            file_extension,
        )
        msg.send()



from datetime import date

from django.db.models.signals import post_save
from django.dispatch import receiver
from dynamic_preferences.registries import global_preferences_registry
from wbcrm.models.accounts import Account
from wbportfolio.models import Claim

from .rebate import manage_rebate_as_task


# This file contains the different receiver directly linked to wbportfolio. In a future iteration, this will be moved outside of wbcommission and we will use a generic commission signal
@receiver(post_save, sender="wbportfolio.Claim")
def post_claim_save_for_rebate_computation(sender, instance, created, **kwargs):
    # if a new commission line is created, we create a general rule
    if instance.status == Claim.Status.APPROVED and instance.account and (root_account := instance.account.get_root()):
        if isinstance(
            instance.date, str
        ):  # we need to do this in case claim are created manually with date as string (allowed). This corner case led to date being still a string when it hits this signal
            instance.refresh_from_db()
        manage_rebate_as_task.delay(
            root_account.id,
            start_date=instance.date,
            only_content_object_ids=[instance.product.id],
            terminal_account_id=instance.account.id,
        )


@receiver(post_save, sender="wbportfolio.Fees")
def post_fees_save_for_rebate_computation(sender, instance, created, **kwargs):
    # if a new commission line is created, we create a general rule
    if (
        created
        and (date.today() - instance.transaction_date).days
        <= global_preferences_registry.manager()["wbcommission__days_to_recompute_rebate_from_fees_threshold"]
    ):  # we make sure that the fee won't trigger rebate computation if they are created too much in the past
        for root_account in Account.objects.filter(level=0):
            if Claim.objects.filter(
                account__in=root_account.get_descendants(include_self=True), product=instance.linked_product
            ).exists():
                manage_rebate_as_task.delay(
                    root_account.id,
                    start_date=instance.transaction_date,
                    only_content_object_ids=[instance.linked_product.id],
                )
